//********************************************************************************************************************************
//*                                                                                                                              *
//*     TMultiThread 1.0.6.2  3delite 2011                                                                                      *
//*     See TMultiThread ReadMe.txt for details                                                                                  *
//*                                                                                                                              *
//* Usage of the component in free programs is FREE.                                                                             *
//* Two licenses are available if you use this component in commercial programs:                                                 *
//* Unlimited Shareware License: 25 Euros                                                                                        *
//* Single Commercial License: 100 Euros                                                                                         *
//*                                                                                                                              *
//*     http://www.shareit.com/product.html?productid=300432175                                                                  *
//*                                                                                                                              *
//*     http://www.3delite.hu/Object Pascal Developer Resources/TMultiThread.html                                                *
//*                                                                                                                              *
//* For other Delphi components see the home page:                                                                               *
//*                                                                                                                              *
//*     http://www.3delite.hu/                                                                                                   *
//*                                                                                                                              *
//* If you have any questions or enquiries please mail: 3delite@3delite.hu                                                       *
//*                                                                                                                              *
//* Good coding! :)                                                                                                              *
//* 3delite                                                                                                                      *
//********************************************************************************************************************************


unit MultiThread;

interface

Uses
    Classes,
    Windows,
    SysUtils,
    Messages;

Const
    MSG_MULTITHREAD_THREAD_FINISHED             = WM_USER + 1;
    MSG_MULTITHREAD_SENT_RESULTS                = WM_USER + 2;
    MSG_MULTITHREAD_TERMINATE_CONTROL_THREAD    = WM_USER + 3;
    MSG_MULTITHREAD_TERMINATE_ALL_THREADS       = WM_USER + 4;
    MSG_MULTITHREAD_ADD_WORK                    = WM_USER + 5;
    MSG_MULTITHREAD_MESSAGE                     = WM_USER + 6;

type
    TMultiThreadThread = class;

    TThreadWorkerCallback = procedure(Thread: TMultiThreadThread; Parameters: TObject; var Data: TObject; var Results: TObject) of object;
    TThreadResultsCallback = procedure(Sender: TObject; var Data: TObject; var Results: TObject) of object;
    TThreadMessage = procedure(Sender: TObject; var Message1, Message2: TObject) of object;

    TThreadCount = (tcAutomatic, tcManual);

    TMultiThreadThreadSettings = class
        ControlThreadHandle: HWND;
        MultiThreadHandle: HWND;
        ThreadWorkerCallback: TThreadWorkerCallback;
        NumberOfThreads: Integer;
        ThreadPriority: TThreadPriority;
        AutoFreeData: Boolean;
        AutoFreeResults: Boolean;
    end;

    TMultiThreadThread = class(TThread)
    private
        procedure SendResults;
    public
        Data: TObject;
        Parameters: TObject;
        Settings: TMultiThreadThreadSettings;
        Results: TObject;
        Constructor Create(Settings: TMultiThreadThreadSettings; Parameters, Data: TObject);
        Destructor Destroy; override;
        procedure SendMessage(Message1, Message2: TObject);
        property Terminated;
    protected
        procedure Execute; override;
    end;

    TMultiThreadControlThread = class(TThread)
    private
        FHWnd: HWND;
        FParameters: TObject;
        FActiveThreads: Integer;
        FDataList: TThreadList;
        FThreadList: TThreadList;
        Settings: TMultiThreadThreadSettings;
        procedure TerminateControlThread;
        procedure TerminateAllThreads;
        procedure AddWork(Data: TObject);
        procedure ThreadFinsihed(Thread: TMultiThreadThread);
        procedure CheckToStartNewThread;
    public
        Constructor Create(Parameters: TObject; MultiThreadThreadSettings: TMultiThreadThreadSettings);
        Destructor Destroy; override;
        procedure Clear;
    //published
        property HWnd: HWND read FHWnd;
        property ActiveThreads: Integer read FActiveThreads;
        property DataList: TThreadList read FDataList;
        property ThreadList: TThreadList read FThreadList;
    protected
        procedure WndMethod(var Msg: TMessage);
        procedure Execute; override;
    end;

    TMultiThread = class(TComponent)
    private
        FHWnd: HWND;
        FThreadCount: TThreadCount;
        FThreadCountManual: Integer;
        FThreadPriority: TThreadPriority;
        FAutoFreeData: Boolean;
        FAutoFreeResults: Boolean;
        FAutoFreeMessages: Boolean;
        FAutoFreeParameters: Boolean;
        EThreadWorkerCallback: TThreadWorkerCallback;
        EThreadResultsCallback: TThreadResultsCallback;
        EThreadMessage: TThreadMessage;
        FMultiThreadControlThread: TMultiThreadControlThread;
        procedure ResultsArrived(Data: TObject; Results: TObject);
        procedure MessageArrived(Message1, Message2: TObject);
    public
        Parameters: TObject;
        Constructor Create(AOwner: TComponent); override;
        Destructor Destroy; override;
        procedure Start;
        procedure Stop;
        procedure StopAllThreads;
        function AddWork(Data: TObject): Boolean;
        function Active: Boolean;
        property HWnd: HWND read FHWnd;
        property MultiThreadControlThread: TMultiThreadControlThread read FMultiThreadControlThread;
    published
        property ThreadCount: TThreadCount read FThreadCount write FThreadCount default tcAutomatic;
        property ThreadCountManual: Integer read FThreadCountManual write FThreadCountManual default 4;
        property ThreadPriority: TThreadPriority read FThreadPriority write FThreadPriority default tpNormal;
        property AutoFreeData: Boolean read FAutoFreeData write FAutoFreeData default True;
        property AutoFreeResults: Boolean read FAutoFreeResults write FAutoFreeResults default True;
        property AutoFreeMessages: Boolean read FAutoFreeMessages write FAutoFreeMessages default True;
        property AutoFreeParameters: Boolean read FAutoFreeParameters write FAutoFreeParameters default True;
        property ThreadWorkerCallback: TThreadWorkerCallback read EThreadWorkerCallback write EThreadWorkerCallback;
        property ThreadResultsCallback: TThreadResultsCallback read EThreadResultsCallback write EThreadResultsCallback;
        property ThreadMessage: TThreadMessage read EThreadMessage write EThreadMessage;
    protected
        procedure WndMethod(var Msg : TMessage);
    end;

    procedure Register;

implementation

procedure Register;
begin
    RegisterComponents('3delite', [TMultiThread]);
end;

Constructor TMultiThreadThread.Create(Settings: TMultiThreadThreadSettings; Parameters, Data: TObject);
begin
    FreeOnTerminate := True;
    Inherited Create(True);
    Priority := Settings.ThreadPriority;
    Self.Settings := Settings;
    Self.Parameters := Parameters;
    Self.Data := Data;
end;

Destructor TMultiThreadThread.Destroy;
begin
    FreeAndNil(Settings);
    Inherited Destroy;
end;

procedure TMultiThreadThread.Execute;
begin
    if NOT Terminated then begin
        if Assigned(Settings.ThreadWorkerCallback) then begin
            try
                Settings.ThreadWorkerCallback(Self, Parameters, Data, Results);
            except
                //*
            end;
        end;
    end;
    if NOT Terminated then begin
        SendResults;
    end else begin
        if Settings.AutoFreeData then begin
            if Assigned(Data) then begin
                FreeAndNil(Data);
            end;
        end;
        if Settings.AutoFreeResults then begin
            if Assigned(Results) then begin
                FreeAndNil(Results);
            end;
        end;
    end;
    PostMessage(Settings.ControlThreadHandle, MSG_MULTITHREAD_THREAD_FINISHED, wParam(Pointer(Self)), 0);
end;

procedure TMultiThreadThread.SendResults;
begin
    PostMessage(Settings.MultiThreadHandle, MSG_MULTITHREAD_SENT_RESULTS, wParam(Pointer(Data)), wParam(Pointer(Results)));
end;

procedure TMultiThreadThread.SendMessage(Message1, Message2: TObject);
begin
    PostMessage(Settings.MultiThreadHandle, MSG_MULTITHREAD_MESSAGE, wParam(Pointer(Message1)), wParam(Pointer(Message2)));
end;

Constructor TMultiThreadControlThread.Create(Parameters: TObject; MultiThreadThreadSettings: TMultiThreadThreadSettings);
begin
    FreeOnTerminate := True;
    Inherited Create(True);
    FHWnd := AllocateHWnd(WndMethod);
    FActiveThreads := 0;
    FDataList := TThreadList.Create;
    FDataList.Duplicates := dupAccept;
    FThreadList := TThreadList.Create;
    FThreadList.Duplicates := dupAccept;
    Settings := MultiThreadThreadSettings;
    FParameters := Parameters;
end;

Destructor TMultiThreadControlThread.Destroy;
begin
    TerminateControlThread;
    Clear;
    FreeAndNil(FDataList);
    FreeAndNil(Settings);
    //DeallocateHWnd(HWnd);
    inherited Destroy;
end;

procedure TMultiThreadControlThread.Execute;
begin
    Repeat
        Sleep(10);
        CheckToStartNewThread;
    Until Terminated;
end;

procedure TMultiThreadControlThread.WndMethod(var Msg : TMessage);
var
    Handled: Boolean;
begin
    Handled := True;
    case Msg.Msg of
        MSG_MULTITHREAD_ADD_WORK: AddWork(TObject(Msg.WParam));
        MSG_MULTITHREAD_THREAD_FINISHED: ThreadFinsihed(TMultiThreadThread(Msg.WParam));
        MSG_MULTITHREAD_TERMINATE_CONTROL_THREAD: TerminateControlThread;
        MSG_MULTITHREAD_TERMINATE_ALL_THREADS: TerminateAllThreads;
    else
        Handled := False;
    end;
    if Handled then begin
        Msg.Result := 0
    end else begin
        Msg.Result := DefWindowProc(FHWnd, Msg.Msg, Msg.WParam, Msg.LParam);
    end;
end;

procedure TMultiThreadControlThread.TerminateControlThread;
begin
    TerminateAllThreads;
    DeallocateHWnd(FHWnd);
    Terminate;
end;

procedure TMultiThreadControlThread.TerminateAllThreads;
var
    i: Integer;
    ThreadHandles: Array of THandle;
    ThreadExitCode: Cardinal;
    List: TList;
begin
    List := FThreadList.LockList;
    SetLength(ThreadHandles, List.Count);
    for i := 0 to List.Count - 1 do begin
        try
            ThreadHandles[i] := TMultiThreadThread(List.Items[i]).Handle;
            GetExitCodeThread(ThreadHandles[i], ThreadExitCode);
            if ThreadExitCode = STILL_ACTIVE then begin
                TMultiThreadThread(List.Items[i]).Terminate;
            end;
        except
            //*
        end;
    end;
    FThreadList.UnlockList;
    Clear;
    WaitForMultipleObjects(Length(ThreadHandles), @ThreadHandles, True, INFINITE);
end;

procedure TMultiThreadControlThread.Clear;
var
    Data: TObject;
    i: Integer;
    List: TList;
begin
    FThreadList.LockList;
    FThreadList.Clear;
    FThreadList.UnlockList;
    List := FDataList.LockList;
    for i := 0 to List.Count - 1 do begin
        Data := TObject(List.Items[i]);
        FreeAndNil(Data);
    end;
    FDataList.UnlockList;
    FDataList.Clear;
end;

procedure TMultiThreadControlThread.AddWork(Data: TObject);
begin
    FDataList.LockList;
    FDataList.Add(Pointer(Data));
    FDataList.UnlockList;
    CheckToStartNewThread;
end;

procedure TMultiThreadControlThread.ThreadFinsihed(Thread: TMultiThreadThread);
var
    Index: Integer;
    List: TList;
begin
    List := FThreadList.LockList;
    Index := List.IndexOf(Pointer(Thread));
    if Index > -1 then begin
        List.Delete(Index);
        List.Pack;
    end;
    FThreadList.UnlockList;
    Dec(FActiveThreads);
    CheckToStartNewThread;
end;

procedure TMultiThreadControlThread.CheckToStartNewThread;
var
    Thread: TMultiThreadThread;
    Data: TObject;
    ThreadSettings: TMultiThreadThreadSettings;
    List: TList;
begin
    if FActiveThreads < Settings.NumberOfThreads then begin
        List := FDataList.LockList;
        if List.Count > 0 then begin
            Data := TObject(List.Items[0]);
            List.Delete(0);
            ThreadSettings := TMultiThreadThreadSettings.Create;
            ThreadSettings.ControlThreadHandle := FHWnd;
            ThreadSettings.ThreadPriority := Settings.ThreadPriority;
            ThreadSettings.MultiThreadHandle := Settings.MultiThreadHandle;
            ThreadSettings.ThreadWorkerCallback := Settings.ThreadWorkerCallback;
            ThreadSettings.AutoFreeData := Settings.AutoFreeData;
            ThreadSettings.AutoFreeResults := Settings.AutoFreeResults;
            Thread := TMultiThreadThread.Create(ThreadSettings, FParameters, Data);
            Inc(FActiveThreads);
            Thread.Resume;
            FThreadList.Add(Pointer(Thread));
        end;
        FDataList.UnlockList;
    end;
end;

procedure TMultiThread.WndMethod(var Msg : TMessage);
var
    Handled: Boolean;
begin
    Handled := True;
    case Msg.Msg of
        MSG_MULTITHREAD_SENT_RESULTS: ResultsArrived(TObject(Msg.WParam), TObject(Msg.LParam));
        MSG_MULTITHREAD_MESSAGE: MessageArrived(TObject(Msg.WParam), TObject(Msg.LParam));
    else
        Handled := False;
    end;
    if Handled then begin
        Msg.Result := 0
    end else begin
        Msg.Result := DefWindowProc(FHWnd, Msg.Msg, Msg.WParam, Msg.LParam);
    end;
end;

procedure TMultiThread.Start;
var
    Settings: TMultiThreadThreadSettings;
    SystemInfo: TSystemInfo;
begin
    Settings := TMultiThreadThreadSettings.Create;
    Settings.MultiThreadHandle := FHWnd;
    Settings.ThreadWorkerCallback := EThreadWorkerCallback;
    if FThreadCount = tcAutomatic then begin
        GetSystemInfo(SystemInfo);
        Settings.NumberOfThreads := SystemInfo.dwNumberOfProcessors * 2;
    end else begin
        Settings.NumberOfThreads := FThreadCountManual;
    end;
    Settings.ThreadPriority := FThreadPriority;
    Settings.AutoFreeData := FAutoFreeData;
    Settings.AutoFreeResults := FAutoFreeResults;
    FMultiThreadControlThread := TMultiThreadControlThread.Create(Parameters, Settings);
    FMultiThreadControlThread.Resume;
end;

procedure TMultiThread.Stop;
var
    ThreadHandles: Array of THandle;
begin
    if FMultiThreadControlThread = nil then begin
        Exit;
    end;
    SetLength(ThreadHandles, 1);
    PostMessage(FMultiThreadControlThread.HWnd, MSG_MULTITHREAD_TERMINATE_CONTROL_THREAD, 0, 0);
    ThreadHandles[0] := FMultiThreadControlThread.Handle;
    WaitForMultipleObjects(1, @ThreadHandles, True, INFINITE);
    FMultiThreadControlThread := nil;
end;

procedure TMultiThread.StopAllThreads;
begin
    if FMultiThreadControlThread = nil then begin
        Exit;
    end;
    PostMessage(FMultiThreadControlThread.HWnd, MSG_MULTITHREAD_TERMINATE_ALL_THREADS, 0, 0);
end;

Constructor TMultiThread.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
    FHWnd := AllocateHWnd(WndMethod);
    FThreadPriority := tpNormal;
    FThreadCountManual := 4;
    FAutoFreeData := True;
    FAutoFreeResults := True;
    FAutoFreeMessages := True;
    FAutoFreeParameters := True;
end;

Destructor TMultiThread.Destroy;
begin
    Stop;
    DeallocateHWnd(FHWnd);
    if FAutoFreeParameters then begin
        if Assigned(Parameters) then begin
            FreeAndNil(Parameters);
        end;
    end;
    inherited Destroy;
end;

procedure TMultiThread.ResultsArrived(Data: TObject; Results: TObject);
begin
    if Assigned(EThreadResultsCallback) then begin
        EThreadResultsCallback(Self, Data, Results);
    end;
    if FAutoFreeData then begin
        if Assigned(Data) then begin
            FreeAndNil(Data);
        end;
    end;
    if FAutoFreeResults then begin
        if Assigned(Results) then begin
            FreeAndNil(Results);
        end;
    end;
end;

procedure TMultiThread.MessageArrived(Message1, Message2: TObject);
begin
    if Assigned(EThreadMessage) then begin
        EThreadMessage(Self, Message1, Message2);
    end;
    if FAutoFreeMessages then begin
        if Assigned(Message1) then begin
            FreeAndNil(Message1);
        end;
        if Assigned(Message2) then begin
            FreeAndNil(Message2);
        end;
    end;
end;

function TMultiThread.AddWork(Data: TObject): Boolean;
begin
    Result := False;
    if FMultiThreadControlThread = nil then begin
        raise Exception.Create('TMultiThread error: can not add job. Control thread not started yet!');
    end;
    PostMessage(FMultiThreadControlThread.HWnd, MSG_MULTITHREAD_ADD_WORK, wParam(Pointer(Data)), 0);
    Result := True;
end;

function TMultiThread.Active: Boolean;
begin
    Result := FMultiThreadControlThread <> nil;
end;

end.
