//********************************************************************************************************************************
//*                                                                                                                              *
//*     TMultiThread 1.0.5 tutorial program  3delite 2011                                                                       *
//*     See TMultiThread ReadMe.txt for details                                                                                  *
//*                                                                                                                              *
//* Usage of the component in free programs is FREE.                                                                             *
//* Two licenses are available if you use this component in commercial programs:                                                 *
//* Unlimited Shareware License: 25 Euros                                                                                        *
//* Single Commercial License: 100 Euros                                                                                         *
//*                                                                                                                              *
//*     http://www.shareit.com/product.html?productid=300432175                                                                  *
//*                                                                                                                              *
//*     http://www.3delite.hu/Object Pascal Developer Resources/TMultiThread.html                                                *
//*                                                                                                                              *
//* For other Delphi components see the home page:                                                                               *
//*                                                                                                                              *
//*     http://www.3delite.hu/                                                                                                   *
//*                                                                                                                              *
//* If you have any questions or enquiries please mail: 3delite@3delite.hu                                                       *
//*                                                                                                                              *
//* Good coding! :)                                                                                                              *
//* 3delite                                                                                                                      *
//********************************************************************************************************************************

unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, MultiThread, StdCtrls;

type
  TForm1 = class(TForm)
    MultiThread1: TMultiThread;
    Button1: TButton;
    Button2: TButton;
    Button3: TButton;
    Memo1: TMemo;
    Button4: TButton;
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure MultiThread1ThreadResultsCallback(Sender: TObject; var Data, Results: TObject);
    procedure MultiThread1ThreadMessage(Sender: TObject; var Message1, Message2: TObject);
    procedure Button4Click(Sender: TObject);
    procedure MultiThread1ThreadWorkerCallback(Thread: TMultiThreadThread; Parameters: TObject; var Data, Results: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

//* We define a class that will hold the information regarding work to do in the thread
//* Here we have 2 data, a work ID and a value to wait in the thread
type
    TWork = class
        ID: Integer;
        TimeToWait: Integer;
    end;

//* Define a class which will transmit our messages from the thread
type
    TWorkMessage = class
        ID: Integer;
        Text: String;
    end;

var
    Form1: TForm1;
    ThreadCounter: Integer = 0;

implementation

{$R *.dfm}

procedure TForm1.Button1Click(Sender: TObject);
begin
    //* Before starting the TMultiThread component you can give it a global Parameters object that
    //* all threads will get, it will be freed automatically when you free the TMultiThread object.
    // MultiThread1.Parameters := OurParameters;
    //* Normally we start the TMultiThread component by code when needed
    MultiThread1.Start;
    Memo1.Lines.Append('TMultiThread listening, waiting for work...');
end;

procedure TForm1.Button2Click(Sender: TObject);
begin
    //* After calling Stop no new threads will be started, already running threads will continue to run
    MultiThread1.Stop;
    Memo1.Lines.Append('TMultiThread stoped.');
end;

procedure TForm1.Button3Click(Sender: TObject);
var
    Work: TWork;
    i: Integer;
begin
    //* Here we create 10 instances of our data class and send them to the TMultiThread component
    //* The TMultiThread component will automatically start the threads, if set ThreadCount to tcAutomatic
    //* then number of CPU cores * 2 threads will work at once maximally
    //* If set to tcManual then set ThreadCountManual to a desired value
    for i := 0 to 9 do begin
        Work := TWork.Create;
        Work.ID := ThreadCounter;
        Work.TimeToWait := 2000;
        MultiThread1.AddWork(Work);
        Inc(ThreadCounter);
    end;
end;

procedure TForm1.Button4Click(Sender: TObject);
begin
    //* Stop all spawned threads and clear the job list
    MultiThread1.StopAllThreads;
end;

procedure TForm1.MultiThread1ThreadWorkerCallback(Thread: TMultiThreadThread; Parameters: TObject; var Data, Results: TObject);
var
    WorkMessage: TWorkMessage;
begin
    //* Here we specify the work to do in the thread
    //* IMPORTANT! DO NOT access VCL stuff here! The code here runs in the thread and VCL is not thread safe!
    //* First we send a message to the TMultiThread component to demonstrate how to send a message
    WorkMessage := TWorkMessage.Create;
    WorkMessage.ID := TWork(Data).ID;
    WorkMessage.Text := 'Starting.';
    //* The "Thread" variable is a TMultiThreadThread class, which has a function to send a message
    //* The function has two parameters, we use the first to send a message
    Thread.SendMessage(WorkMessage, nil);
    //* As a thread work for demonstration we wait for the given time
    Sleep(TWork(Data).TimeToWait);
    //* If the thread is terminated in the mean time we can be notifyied like this:
    if Thread.Terminated then begin
        //* For demonstration we also send a message about this event
        WorkMessage := TWorkMessage.Create;
        WorkMessage.ID := TWork(Data).ID;
        WorkMessage.Text := 'Terminated.';
        Thread.SendMessage(WorkMessage, nil);
    end;
    //* To send back the result of the work create a new object and assign it to Results,
    //* In the results callback you will get this object back.
end;

procedure TForm1.MultiThread1ThreadMessage(Sender: TObject; var Message1, Message2: TObject);
begin
    //* Message1 contains our TWorkMessage class, our message from the thread
    //* Type-cast our Message1 variable to our TworkMessage class to get the message,
    //* and write out the message as we are in the main thread
    Memo1.Lines.Append('Thread message arrived from thread: '
        + IntToStr(TWorkMessage(Message1).ID) + ': '
        + TWorkMessage(Message1).Text);
end;

procedure TForm1.MultiThread1ThreadResultsCallback(Sender: TObject; var Data, Results: TObject);
begin
    //* We get the result here, we are in the main thread, VCL access is available
    //* For demonstration we just display the thread ID that has finished.
    Memo1.Lines.Append('Thread ' + IntToStr(TWork(Data).ID) + ' finished.');
end;

end.
